@@ -13,10 +13,13 @@
 #include <unistd.h>
 
 #include "base/error_state.h"
+#include "base/error_string.h"
 #include "base/nested_output.h"
+#include "caching/commands.h"
 #include "caching/metadata.h"
 #include "persistent-data/block.h"
 #include "persistent-data/file_utils.h"
+#include "persistent-data/space_map.h"
 #include "persistent-data/space-maps/core.h"
 #include "version.h"
 
@@ -158,6 +161,15 @@ namespace {
 		using reporter_base::get_error;
 	};
 
+	class space_map_reporter : public space_map_detail::visitor, reporter_base {
+	public:
+		space_map_reporter(nested_output &o)
+		: reporter_base(o) {
+		}
+
+		using reporter_base::get_error;
+	};
+
 	//--------------------------------
 
 	transaction_manager::ptr open_tm(block_manager<>::ptr bm) {
@@ -175,7 +187,8 @@ namespace {
 			  check_hints_(true),
 			  check_discards_(true),
 			  ignore_non_fatal_errors_(false),
-			  quiet_(false) {
+			  quiet_(false),
+			  clear_needs_check_on_success_(false) {
 		}
 
 		bool check_mappings_;
@@ -183,6 +196,7 @@ namespace {
 		bool check_discards_;
 		bool ignore_non_fatal_errors_;
 		bool quiet_;
+		bool clear_needs_check_on_success_;
 	};
 
 	struct stat guarded_stat(string const &path) {
@@ -191,16 +205,21 @@ namespace {
 		int r = ::stat(path.c_str(), &info);
 		if (r) {
 			ostringstream msg;
-			char buffer[128], *ptr;
-
-			ptr = ::strerror_r(errno, buffer, sizeof(buffer));
-			msg << path << ": " << ptr;
+			msg << path << ": " << error_string(errno);
 			throw runtime_error(msg.str());
 		}
 
 		return info;
 	}
 
+	void clear_needs_check(string const &path) {
+		block_manager<>::ptr bm = open_bm(path, block_manager<>::READ_WRITE);
+
+		superblock sb = read_superblock(bm);
+		sb.flags.clear_flag(superblock_flags::NEEDS_CHECK);
+		write_superblock(bm, sb);
+	}
+
 	error_state metadata_check(block_manager<>::ptr bm, flags const &fs) {
 		nested_output out(cerr, 2);
 		if (fs.quiet_)
@@ -227,7 +246,7 @@ namespace {
 			out << "examining mapping array" << end_message();
 			{
 				nested_output::nest _ = out.push();
-				mapping_array ma(tm, mapping_array::ref_counter(), sb.mapping_root, sb.cache_blocks);
+				mapping_array ma(*tm, mapping_array::ref_counter(), sb.mapping_root, sb.cache_blocks);
 				check_mapping_array(ma, mapping_rep);
 			}
 		}
@@ -240,7 +259,7 @@ namespace {
 				out << "examining hint array" << end_message();
 				{
 					nested_output::nest _ = out.push();
-					hint_array ha(tm, sb.policy_hint_size, sb.hint_root, sb.cache_blocks);
+					hint_array ha(*tm, sb.policy_hint_size, sb.hint_root, sb.cache_blocks);
 					ha.check(hint_rep);
 				}
 			}
@@ -254,7 +273,7 @@ namespace {
 				out << "examining discard bitset" << end_message();
 				{
 					nested_output::nest _ = out.push();
-					bitset discards(tm, sb.discard_root, sb.discard_nr_blocks);
+					persistent_data::bitset discards(*tm, sb.discard_root, sb.discard_nr_blocks);
 				}
 			}
 		}
@@ -276,9 +295,19 @@ namespace {
 			throw runtime_error(msg.str());
 		}
 
-		block_manager<>::ptr bm = open_bm(path, block_io<>::READ_ONLY);
+		block_manager<>::ptr bm = open_bm(path, block_manager<>::READ_ONLY);
 		err = metadata_check(bm, fs);
 
+		bool success = false;
+
+		if (fs.ignore_non_fatal_errors_)
+			success = (err == FATAL) ? false : true;
+		else
+			success = (err == NO_ERROR) ? true : false;
+
+		if (success && fs.clear_needs_check_on_success_)
+			clear_needs_check(path);
+
 		return err == NO_ERROR ? 0 : 1;
 	}
 
@@ -303,6 +332,7 @@ namespace {
 		    << "  {-q|--quiet}" << endl
 		    << "  {-h|--help}" << endl
 		    << "  {-V|--version}" << endl
+		    << "  {--clear-needs-check-flag}" << endl
 		    << "  {--super-block-only}" << endl
 		    << "  {--skip-mappings}" << endl
 		    << "  {--skip-hints}" << endl
@@ -312,17 +342,18 @@ namespace {
 
 //----------------------------------------------------------------
 
-int main(int argc, char **argv)
+int cache_check_main(int argc, char **argv)
 {
 	int c;
 	flags fs;
 	const char shortopts[] = "qhV";
 	const struct option longopts[] = {
 		{ "quiet", no_argument, NULL, 'q' },
-		{ "superblock-only", no_argument, NULL, 1 },
+		{ "super-block-only", no_argument, NULL, 1 },
 		{ "skip-mappings", no_argument, NULL, 2 },
 		{ "skip-hints", no_argument, NULL, 3 },
 		{ "skip-discards", no_argument, NULL, 4 },
+		{ "clear-needs-check-flag", no_argument, NULL, 5 },
 		{ "help", no_argument, NULL, 'h' },
 		{ "version", no_argument, NULL, 'V' },
 		{ NULL, no_argument, NULL, 0 }
@@ -347,6 +378,10 @@ int main(int argc, char **argv)
 			fs.check_discards_ = false;
 			break;
 
+		case 5:
+			fs.clear_needs_check_on_success_ = true;
+			break;
+
 		case 'h':
 			usage(cout, basename(argv[0]));
 			return 0;
@@ -374,4 +409,6 @@ int main(int argc, char **argv)
 	return check_with_exception_handling(argv[optind], fs);
 }
 
+base::command caching::cache_check_cmd("cache_check", cache_check_main);
+
 //----------------------------------------------------------------
